!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_WARM
INTERFACE
SUBROUTINE ICE4_WARM(LDSOFT, PCOMPUTE, HSUBG_RC_RR_ACCR, HSUBG_RR_EVAP, &
                     PRHODREF, PLVFACT, PT, PPRES, PTHT, &
                     PLBDAR, PLBDAR_RF, PKA, PDV, PCJ, &
                     PHLC_LCF, PHLC_HCF, PHLC_LRC, PHLC_HRC, &
                     PCF, PRF, &
                     PRVT, PRCT, PRRT, &
                     PRCAUTR, PRCACCR, PRREVAV, &
                     PA_TH, PA_RV, PA_RC, PA_RR)
IMPLICIT NONE
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
CHARACTER(len=80),        INTENT(IN)    :: HSUBG_RC_RR_ACCR ! subgrid rc-rr accretion
CHARACTER(len=80),        INTENT(IN)    :: HSUBG_RR_EVAP ! subgrid rr evaporation
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PTHT     ! Theta at time t
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR_RF!like PLBDAR but for the Rain Fraction part
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_HCF ! HLCLOUDS : fraction of High Cloud Fraction in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_LCF ! HLCLOUDS : fraction of Low  Cloud Fraction in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_HRC ! HLCLOUDS : LWC that is High LWC in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_LRC ! HLCLOUDS : LWC that is Low  LWC in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PCF      ! Cloud fraction
REAL, DIMENSION(:),       INTENT(IN)    :: PRF      ! Rain fraction
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCAUTR   ! Autoconversion of r_c for r_r production
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCACCR  ! Accretion of r_c for r_r production
REAL, DIMENSION(:),       INTENT(INOUT) :: PRREVAV  ! Evaporation of r_r
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RV
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
END SUBROUTINE ICE4_WARM
END INTERFACE
END MODULE MODI_ICE4_WARM
SUBROUTINE ICE4_WARM(LDSOFT, PCOMPUTE, HSUBG_RC_RR_ACCR, HSUBG_RR_EVAP, &
                     PRHODREF, PLVFACT, PT, PPRES, PTHT, &
                     PLBDAR, PLBDAR_RF, PKA, PDV, PCJ, &
                     PHLC_LCF, PHLC_HCF, PHLC_LRC, PHLC_HRC, &
                     PCF, PRF, &
                     PRVT, PRCT, PRRT, &
                     PRCAUTR, PRCACCR, PRREVAV, &
                     PA_TH, PA_RV, PA_RC, PA_RR)
!!
!!**  PURPOSE
!!    -------
!!      Computes the warm process
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the plitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XALPW,XBETAW,XCL,XCPD,XCPV,XGAMW,XLVTT,XMD,XMV,XRV,XTT,XEPSILO
USE MODD_RAIN_ICE_DESCR, ONLY: XCEXVT,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: X0EVAR,X1EVAR,XCRIAUTC,XEX0EVAR,XEX1EVAR,XEXCACCR,XFCACCR,XTIMAUTC
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MSG
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
CHARACTER(len=80),        INTENT(IN)    :: HSUBG_RC_RR_ACCR ! subgrid rc-rr accretion
CHARACTER(len=80),        INTENT(IN)    :: HSUBG_RR_EVAP ! subgrid rr evaporation
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PTHT     ! Theta at time t
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR_RF!like PLBDAR but for the Rain Fraction part
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_HCF ! HLCLOUDS : fraction of High Cloud Fraction in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_LCF ! HLCLOUDS : fraction of Low  Cloud Fraction in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_HRC ! HLCLOUDS : LWC that is High LWC in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PHLC_LRC ! HLCLOUDS : LWC that is Low  LWC in grid
REAL, DIMENSION(:),       INTENT(IN)    :: PCF      ! Cloud fraction
REAL, DIMENSION(:),       INTENT(IN)    :: PRF      ! Rain fraction
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCAUTR   ! Autoconversion of r_c for r_r production
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCACCR  ! Accretion of r_c for r_r production
REAL, DIMENSION(:),       INTENT(INOUT) :: PRREVAV  ! Evaporation of r_r
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RV
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
!
!*       0.2  declaration of local variables
!
INTEGER                            :: ISIZE
INTEGER                            :: JL
#ifndef MNH_OPENACC
REAL,    DIMENSION(:), allocatable :: ZZW2, ZZW3, ZZW4
REAL,    DIMENSION(:), allocatable :: ZUSW ! Undersaturation over water
REAL,    DIMENSION(:), allocatable :: ZTHLT    ! Liquid potential temperature
REAL,    DIMENSION(:), allocatable :: ZMASK, ZMASK1, ZMASK2
#else
REAL,    DIMENSION(:), pointer, contiguous :: ZZW2, ZZW3, ZZW4
REAL,    DIMENSION(:), pointer, contiguous :: ZUSW ! Undersaturation over water
REAL,    DIMENSION(:), pointer, contiguous :: ZTHLT    ! Liquid potential temperature
REAL,    DIMENSION(:), pointer, contiguous :: ZMASK, ZMASK1, ZMASK2
#endif

LOGICAL :: GHSUBG_RR_EVAP ! temporary variable for OpenACC character limitation (Cray CCE)

!$acc data present(PCOMPUTE,PRHODREF,PLVFACT,PT,PPRES,PTHT, &
!$acc&                PLBDAR,PLBDAR_RF,PKA,PDV,PCJ,PHLC_LCF,PHLC_HCF,PHLC_LRC,PHLC_HRC,PCF,PRF,             &
!$acc&                PRVT,PRCT,PRRT,PRCAUTR,PRCACCR,PRREVAV,PA_TH,PA_RV,PA_RC,PA_RR                        )
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK1D(PCOMPUTE,"ICE4_WARM beg:PCOMPUTE")
  CALL MPPDB_CHECK1D(PRHODREF,"ICE4_WARM beg:PRHODREF",PRECISION)
  CALL MPPDB_CHECK1D(PLVFACT,"ICE4_WARM beg:PLVFACT",PRECISION)
  CALL MPPDB_CHECK1D(PT,"ICE4_WARM beg:PT",PRECISION)
  CALL MPPDB_CHECK1D(PPRES,"ICE4_WARM beg:PPRES",PRECISION)
  CALL MPPDB_CHECK1D(PTHT,"ICE4_WARM beg:PTHT",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAR,"ICE4_WARM beg:PLBDAR",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAR_RF,"ICE4_WARM beg:PLBDAR_RF",PRECISION)
  CALL MPPDB_CHECK1D(PKA,"ICE4_WARM beg:PKA",PRECISION)
  CALL MPPDB_CHECK1D(PDV,"ICE4_WARM beg:PDV",PRECISION)
  CALL MPPDB_CHECK1D(PCJ,"ICE4_WARM beg:PCJ",PRECISION)
  CALL MPPDB_CHECK1D(PHLC_HCF,"ICE4_WARM beg:PHLC_HCF",PRECISION)
  CALL MPPDB_CHECK1D(PHLC_LCF,"ICE4_WARM beg:PHLC_LCF",PRECISION)
  CALL MPPDB_CHECK1D(PHLC_HRC,"ICE4_WARM beg:PHLC_HRC",PRECISION)
  CALL MPPDB_CHECK1D(PHLC_LRC,"ICE4_WARM beg:PHLC_LRC",PRECISION)
  CALL MPPDB_CHECK1D(PCF,"ICE4_WARM beg:PCF",PRECISION)
  CALL MPPDB_CHECK1D(PRF,"ICE4_WARM beg:PRF",PRECISION)
  CALL MPPDB_CHECK1D(PRVT,"ICE4_WARM beg:PRVT",PRECISION)
  CALL MPPDB_CHECK1D(PRCT,"ICE4_WARM beg:PRCT",PRECISION)
  CALL MPPDB_CHECK1D(PRRT,"ICE4_WARM beg:PRRT",PRECISION)
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRCAUTR,"ICE4_WARM beg:PRCAUTR",PRECISION)
  CALL MPPDB_CHECK1D(PRCACCR,"ICE4_WARM beg:PRCACCR",PRECISION)
  CALL MPPDB_CHECK1D(PRREVAV,"ICE4_WARM beg:PRREVAV",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_WARM beg:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RV,"ICE4_WARM beg:PA_RV",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_WARM beg:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_WARM beg:PA_RR",PRECISION)
END IF

ISIZE = Size( PRHODREF )

#ifndef MNH_OPENACC
allocate( zzw2  (ISIZE) )
allocate( zzw3  (ISIZE) )
allocate( zzw4  (ISIZE) )
allocate( zusw  (ISIZE) )
allocate( zthlt (ISIZE) )
allocate( zmask (ISIZE) )
allocate( zmask1(ISIZE) )
allocate( zmask2(ISIZE) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ICE4_WARM' )

CALL MNH_MEM_GET( zzw2,   isize )
CALL MNH_MEM_GET( zzw3,   isize )
CALL MNH_MEM_GET( zzw4,   isize )
CALL MNH_MEM_GET( zusw,   isize )
CALL MNH_MEM_GET( zthlt,  isize )
CALL MNH_MEM_GET( zmask,  isize )
CALL MNH_MEM_GET( zmask1, isize )
CALL MNH_MEM_GET( zmask2, isize )

!$acc data present( ZZW2, ZZW3, ZZW4, ZUSW, ZTHLT, ZMASK, ZMASK1, ZMASK2 )
#endif

!
!*       4.2    compute the autoconversion of r_c for r_r production: RCAUTR
!
!$acc kernels
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(2)-PHLC_HRC(JL))) * & ! PHLC_HRC(:)>XRTMIN(2)
           &MAX(0., -SIGN(1., 1.E-20-PHLC_HCF(JL))) * & ! PHLC_HCF(:) .GT. 0.
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRCAUTR(JL)=PRCAUTR(JL)*ZMASK(JL)
  ENDDO
ELSE
  PRCAUTR(:) = 0.
  WHERE(ZMASK(:)==1.)
    PRCAUTR(:) = XTIMAUTC*MAX(PHLC_HRC(:)/PHLC_HCF(:) - XCRIAUTC/PRHODREF(:), 0.0)
    PRCAUTR(:) = PHLC_HCF(:)*PRCAUTR(:)
  END WHERE
ENDIF
DO JL=1, ISIZE
  PA_RC(JL) = PA_RC(JL) - PRCAUTR(JL)
  PA_RR(JL) = PA_RR(JL) + PRCAUTR(JL)
ENDDO
!$acc end kernels
!
!
!*       4.3    compute the accretion of r_c for r_r production: RCACCR
!
IF (HSUBG_RC_RR_ACCR=='NONE') THEN
!$acc kernels
  !CLoud water and rain are diluted over the grid box
  DO JL=1, ISIZE
    ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(2)-PRCT(JL))) * & ! PRCT(:)>XRTMIN(2)
             &MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JL))) * & ! PRRT(:)>XRTMIN(3)
             &PCOMPUTE(JL)
  ENDDO
  IF(LDSOFT) THEN
    DO JL=1, ISIZE
      PRCACCR(JL)=PRCACCR(JL) * ZMASK(JL)
    ENDDO
  ELSE
    PRCACCR(:) = 0.
    WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
      PRCACCR(:) = XFCACCR * PRCT(:)                &
                 * PLBDAR(:)**XEXCACCR    &
                 * PRHODREF(:)**(-XCEXVT)
#else
      PRCACCR(:) = XFCACCR * PRCT(:)           &
                 * BR_POW(PLBDAR(:),XEXCACCR)  &
                 * BR_POW(PRHODREF(:),-XCEXVT)
#endif
    END WHERE
  ENDIF
!$acc end kernels

ELSEIF (HSUBG_RC_RR_ACCR=='PRFR') THEN
!$acc kernels
  !Cloud water is concentrated over its fraction with possibly to parts with high and low content as set for autoconversion
  !Rain is concnetrated over its fraction
  !Rain in high content area fraction: PHLC_HCF
  !Rain in low content area fraction:
  ! if PRF<PCF (rain is entirely falling in cloud): PRF-PHLC_HCF
  ! if PRF>PCF (rain is falling in cloud and in clear sky): PCF-PHLC_HCF
  ! => min(PCF, PRF)-PHLC_HCF
  !$mnh_do_concurrent ( JL = 1 : ISIZE )
    ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(2)-PRCT(JL))) * & ! PRCT(:)>XRTMIN(2)
             &MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JL))) * & ! PRRT(:)>XRTMIN(3)
             &PCOMPUTE(JL)
    ZMASK1(JL)=ZMASK(JL) * &
              &MAX(0., -SIGN(1., XRTMIN(2)-PHLC_HRC(JL))) * & ! PHLC_HRC(:)>XRTMIN(2)
              &MAX(0., -SIGN(1., 1.E-20-PHLC_HCF(JL))) ! PHLC_HCF(:)>0.
    ZMASK2(JL)=ZMASK(JL) * &
              &MAX(0., -SIGN(1., XRTMIN(2)-PHLC_LRC(JL))) * & ! PHLC_LRC(:)>XRTMIN(2)
              &MAX(0., -SIGN(1., 1.E-20-PHLC_LCF(JL))) ! PHLC_LCF(:)>0.
  ENDDO
  IF(LDSOFT) THEN
    DO JL=1, ISIZE
      PRCACCR(JL)=PRCACCR(JL) * MIN(1., ZMASK1(JL)+ZMASK2(JL))
    ENDDO
  ELSE
    PRCACCR(:)=0.
    WHERE(ZMASK1(:)==1.)
      !Accretion due to rain falling in high cloud content
#ifndef MNH_BITREP
      PRCACCR(:) = XFCACCR * ( PHLC_HRC(:)/PHLC_HCF(:) )     &
             * PLBDAR_RF(:)**XEXCACCR &
             * PRHODREF(:)**(-XCEXVT) &
             * PHLC_HCF
#else
      PRCACCR(:) = XFCACCR * ( PHLC_HRC(:)/PHLC_HCF(:) ) &
             * BR_POW(PLBDAR_RF(:),XEXCACCR)             &
             * BR_POW(PRHODREF(:),-XCEXVT)               &
             * PHLC_HCF
#endif
    END WHERE
    WHERE(ZMASK2(:)==1.)
      !We add acrretion due to rain falling in low cloud content
#ifndef MNH_BITREP
      PRCACCR(:) = PRCACCR(:) + XFCACCR * ( PHLC_LRC(:)/PHLC_LCF(:) )     &
                      * PLBDAR_RF(:)**XEXCACCR &
                      * PRHODREF(:)**(-XCEXVT) &
                      * (MIN(PCF(:), PRF(:))-PHLC_HCF(:))
#else
      PRCACCR(:) = PRCACCR(:) + XFCACCR * ( PHLC_LRC(:)/PHLC_LCF(:) )  &
                      * BR_POW(PLBDAR_RF(:),XEXCACCR)                  &
                      * BR_POW(PRHODREF(:),-XCEXVT)                    &
                      * (MIN(PCF(:), PRF(:))-PHLC_HCF(:))
#endif
    END WHERE
  ENDIF
!$acc end kernels
ELSE
  CALL PRINT_MSG(NVERB_FATAL,'GEN','ICE4_WARM','wrong HSUBG_RC_RR_ACCR case')
ENDIF
!$acc kernels
DO JL=1, ISIZE
  PA_RC(JL) = PA_RC(JL) - PRCACCR(JL)
  PA_RR(JL) = PA_RR(JL) + PRCACCR(JL)
ENDDO
!$acc end kernels
!
!*       4.4    compute the evaporation of r_r: RREVAV
!
IF (HSUBG_RR_EVAP=='NONE') THEN
!$acc kernels
  DO JL=1, ISIZE
    ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JL))) * & ! PRRT(:)>XRTMIN(3)
             &MAX(0., SIGN(1., XRTMIN(2)-PRCT(JL))) * & ! PRCT(:)<=XRTMIN(2)
             &PCOMPUTE(JL)
  ENDDO
  IF(LDSOFT) THEN
    DO JL=1, ISIZE
      PRREVAV(JL)=PRREVAV(JL)*ZMASK(JL)
    ENDDO
  ELSE
    PRREVAV(:) = 0.
    !Evaporation only when there's no cloud (RC must be 0)
    !$mnh_do_concurrent ( JL = 1 : ISIZE )
      IF ( ZMASK(Jl) == 1. ) THEN
#ifndef MNH_BITREP
        PRREVAV(JL)  = EXP( XALPW - XBETAW/PT(JL) - XGAMW*ALOG(PT(JL) ) ) ! es_w
        ZUSW(JL) = 1.0 - PRVT(JL)*( PPRES(JL)-PRREVAV(JL) ) / ( XEPSILO * PRREVAV(JL) )
                                                      ! Undersaturation over water
        PRREVAV(JL) = ( XLVTT+(XCPV-XCL)*(PT(JL)-XTT) )**2 / ( PKA(JL)*XRV*PT(JL)**2 ) &
             + ( XRV*PT(JL) ) / ( PDV(JL)*PRREVAV(JL) )
        PRREVAV(JL) = ( MAX( 0.0,ZUSW(JL) )/(PRHODREF(JL)*PRREVAV(JL)) ) *      &
          ( X0EVAR*PLBDAR(JL)**XEX0EVAR+X1EVAR*PCJ(JL)*PLBDAR(JL)**XEX1EVAR )
#else
        PRREVAV(JL)  = BR_EXP( XALPW - XBETAW/PT(JL) - XGAMW*BR_LOG(PT(JL) ) ) ! es_w
        ZUSW(JL) = 1.0 - PRVT(JL)*( PPRES(JL)-PRREVAV(JL) ) / ( (XMV/XMD) * PRREVAV(JL) )
                                                      ! Undersaturation over water
        PRREVAV(JL) = BR_P2( XLVTT+(XCPV-XCL)*(PT(JL)-XTT) ) / ( PKA(JL)*XRV*BR_P2(PT(JL)) ) &
             + ( XRV*PT(JL) ) / ( PDV(JL)*PRREVAV(JL) )
        PRREVAV(JL) = ( MAX( 0.0,ZUSW(JL) )/(PRHODREF(JL)*PRREVAV(JL)) ) *      &
          ( X0EVAR*BR_POW(PLBDAR(JL),XEX0EVAR)+X1EVAR*PCJ(JL)*BR_POW(PLBDAR(JL),XEX1EVAR) )
#endif
      END IF
    !$mnh_end_do()
  ENDIF
!$acc end kernels

ELSEIF (HSUBG_RR_EVAP=='CLFR' .OR. HSUBG_RR_EVAP=='PRFR') THEN
GHSUBG_RR_EVAP=.false.
IF (HSUBG_RR_EVAP=='CLFR') GHSUBG_RR_EVAP=.true.
!$acc kernels
  !Evaporation in clear sky part
  !With CLFR, rain is diluted over the grid box
  !With PRFR, rain is concentrated in its fraction
  !Use temperature and humidity in clear sky part like Bechtold et al. (1993)
  !IF (HSUBG_RR_EVAP=='CLFR') THEN
  IF (GHSUBG_RR_EVAP) THEN
    ZZW4(:)=1. !Precipitation fraction
    ZZW3(:)=PLBDAR(:)
  ELSE
    ZZW4(:)=PRF(:) !Precipitation fraction
    ZZW3(:)=PLBDAR_RF(:)
  ENDIF

  !ATTENTION
  !Il faudrait recalculer les variables PKA, PDV, PCJ en tenant compte de la température T^u
  !Ces variables devraient être sorties de rain_ice_slow et on mettrait le calcul de T^u, T^s
  !et plusieurs versions (comme actuellement, en ciel clair, en ciel nuageux) de PKA, PDV, PCJ dans rain_ice
  !On utiliserait la bonne version suivant l'option NONE, CLFR... dans l'évaporation et ailleurs
  !$mnh_do_concurrent ( JL = 1 : ISIZE )
    ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JL))) * & ! PRRT(:)>XRTMIN(3)
             &MAX(0., -SIGN(1., PCF(JL)-ZZW4(JL))) * & ! ZZW4(:) > PCF(:)
             &PCOMPUTE(JL)
  ENDDO
  IF(LDSOFT) THEN
    DO JL=1, ISIZE
      PRREVAV(JL)=PRREVAV(JL)*ZMASK(JL)
    ENDDO
  ELSE
    PRREVAV(:) = 0.
    !$mnh_do_concurrent ( JL = 1 : ISIZE )
      IF ( ZMASK(Jl) == 1. ) THEN
        ! outside the cloud (environment) the use of T^u (unsaturated) instead of T
        ! Bechtold et al. 1993
        !
        ! T_l
        ZTHLT(JL) = PTHT(JL) - XLVTT*PTHT(JL)/XCPD/PT(JL)*PRCT(JL)
        !
        ! T^u = T_l = theta_l * (T/theta)
        ZZW2(JL) =  ZTHLT(JL) * PT(JL) / PTHT(JL)
        !
        ! es_w with new T^u
#ifndef MNH_BITREP
        PRREVAV(JL)  = EXP( XALPW - XBETAW/ZZW2(JL) - XGAMW*ALOG(ZZW2(JL) ) )
#else
        PRREVAV(JL)  = BR_EXP( XALPW - XBETAW/ZZW2(JL) - XGAMW*BR_LOG(ZZW2(JL) ) )
#endif
        !
        ! S, Undersaturation over water (with new theta^u)
        ZUSW(JL) = 1.0 - PRVT(JL)*( PPRES(JL)-PRREVAV(JL) ) / ( XEPSILO * PRREVAV(JL) )
        !
#ifndef MNH_BITREP
        PRREVAV(JL) = ( XLVTT+(XCPV-XCL)*(ZZW2(JL)-XTT) )**2 / ( PKA(JL)*XRV*ZZW2(JL)**2 ) &
               + ( XRV*ZZW2(JL) ) / ( PDV(JL)*PRREVAV(JL) )
        !
        PRREVAV(JL) = MAX( 0.0,ZUSW(JL) )/(PRHODREF(JL)*PRREVAV(JL))  *      &
               ( X0EVAR*ZZW3(JL)**XEX0EVAR+X1EVAR*PCJ(JL)*ZZW3(JL)**XEX1EVAR )
#else
        PRREVAV(JL) = BR_P2( XLVTT+(XCPV-XCL)*(ZZW2(JL)-XTT) ) / ( PKA(JL)*XRV*BR_P2(ZZW2(JL)) ) &
               + ( XRV*ZZW2(JL) ) / ( PDV(JL)*PRREVAV(JL) )
        !
        PRREVAV(JL) = MAX( 0.0,ZUSW(JL) )/(PRHODREF(JL)*PRREVAV(JL))  *                     &
               ( X0EVAR*BR_POW(ZZW3(JL),XEX0EVAR)+X1EVAR*PCJ(JL)*BR_POW(ZZW3(JL),XEX1EVAR) )
#endif
        !
        PRREVAV(JL) = PRREVAV(JL)*(ZZW4(JL)-PCF(JL))
      END IF
    !$mnh_end_do()
  ENDIF
!$acc end kernels

ELSE
  CALL PRINT_MSG(NVERB_FATAL,'GEN','ICE4_WARM','wrong HSUBG_RR_EVAP case')
END IF
!$acc kernels
DO JL=1, ISIZE
  PA_RR(JL) = PA_RR(JL) - PRREVAV(JL)
  PA_RV(JL) = PA_RV(JL) + PRREVAV(JL)
  PA_TH(JL) = PA_TH(JL) - PRREVAV(JL)*PLVFACT(JL)
ENDDO
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRCAUTR,"ICE4_WARM end:PRCAUTR",PRECISION)
  CALL MPPDB_CHECK1D(PRCACCR,"ICE4_WARM end:PRCACCR",PRECISION)
  CALL MPPDB_CHECK1D(PRREVAV,"ICE4_WARM end:PRREVAV",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_WARM end:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RV,"ICE4_WARM end:PA_RV",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_WARM end:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_WARM end:PA_RR",PRECISION)
END IF

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'ICE4_WARM' )
#endif

!$acc end data

END SUBROUTINE ICE4_WARM

